package data.scripts.shipai;

import com.fs.starfarer.api.Global;
import com.fs.starfarer.api.combat.DroneLauncherShipSystemAPI;
import com.fs.starfarer.api.combat.DroneLauncherShipSystemAPI.DroneOrders;
import com.fs.starfarer.api.combat.FluxTrackerAPI;
import com.fs.starfarer.api.combat.ShipAIPlugin;
import com.fs.starfarer.api.combat.ShipAPI;
import com.fs.starfarer.api.combat.ShipCommand;
import com.fs.starfarer.api.combat.ShipSystemAIScript;
import com.fs.starfarer.api.combat.ShipwideAIFlags;
import com.fs.starfarer.api.util.IntervalUtil;

import static data.scripts.shipai.AutoFireController.AutoFire;
import static data.scripts.shipai.ShieldManager.ShieldManager;
import static data.scripts.shipai.TargetManager.TargetManager;
import static data.scripts.shipai.SteerMeStandard.getAngleDifference;

import org.lazywizard.lazylib.MathUtils;
import org.lazywizard.lazylib.VectorUtils;

import org.lwjgl.util.vector.Vector2f;
import static data.scripts.shipai.BestStrafeDistance.BestStrafeDistance;

//Drone AI
//Attempts to stay near the mother ship and attack nearby foes.
public class StandardDroneAI implements ShipAIPlugin
{
    // Our ship object
    private final ShipAPI ship;
    private final ShipAPI motherShip;
    private final DroneLauncherShipSystemAPI droneSystem;

    //Our System AI
    private final ShipSystemAIScript script;
    //Phase Cloak uses its own sub-AI for convenience.
    private final ShipSystemAIScript cloakScript;

    // Our current target (can be null)
    private ShipAPI target = null;

    //Booleans set current control state, updated every tick
    private boolean isAccelerating = false;
    private boolean stopAccelerating = false;
    private boolean strafeRight = false;
    private boolean strafeLeft = false;
    //Used for steering decisions
    private Vector2f destination = null;

    private final IntervalUtil checkInterval = new IntervalUtil(0.05f, 0.15f);
    //longInterval causes a new evalutation of targets regardless of anything else, to prevent AIs getting stuck stupidly.
    //Warning; don't run this too often, otherwise the AI won't even aim the guns properly.
    private final IntervalUtil longInterval = new IntervalUtil(2.5f, 8.0f);

    private boolean shouldStrafe;
    private float strafeDistance;
    private final float closeDistance;
    private final float roamDistance;
    private boolean canCircleStrafe = false;
    private boolean strafeState = true;
    private final IntervalUtil strafeInterval = new IntervalUtil(0.5f,1f);
	private final ShipwideAIFlags flags = new ShipwideAIFlags();


    //Initial setup.  If waitTime = 0f, the guidance begins one frame after launch.
    //Otherwise, drifts until waitTime = 0f (useful for anime-style launch systems)
    public StandardDroneAI (ShipAPI theShip, ShipAPI mothership, DroneLauncherShipSystemAPI dronesystem, ShipSystemAIScript theScript, ShipSystemAIScript theCloakScript, float closedistance, float roamdistance,  boolean theShouldStrafe, float theStrafeDistance)
    {
        this.ship = theShip;
        this.motherShip = mothership;
        this.droneSystem = dronesystem;
        this.closeDistance = closedistance;
        this.roamDistance = roamdistance;
        this.shouldStrafe = theShouldStrafe;
        this.strafeDistance = theStrafeDistance;
        this.script = theScript;
        if(script != null && ship.getSystem() != null){
            script.init(ship,ship.getSystem(),null,Global.getCombatEngine());
        }
        this.cloakScript = theCloakScript;
        if(cloakScript != null && ship.getPhaseCloak() != null){
            cloakScript.init(ship,ship.getPhaseCloak(),null,Global.getCombatEngine());
        }
    }

    //Main logic
    @Override
    public void advance(float amount)
    {
        //LOGIC CHECKS
        if(Global.getCombatEngine() == null) return;
        if (Global.getCombatEngine().isPaused()) return;


        //If the engine's not paused, tick tock
        checkInterval.advance(amount);
        longInterval.advance(amount);
        strafeInterval.advance(amount);

        //Automatically strafe back and forth when that's vaguely appropriate.
        if(strafeInterval.intervalElapsed()){
            if(canCircleStrafe && shouldStrafe){
                strafeState = !strafeState;
                if(strafeState){
                    strafeRight = true;
                    strafeLeft = false;
                    strafeInterval.randomize();
                }else{
                    strafeRight = false;
                    strafeLeft = true;
                    strafeInterval.randomize();
                }
            }
        }

        //If enough time has passed, re-evaluate our best target, reset our weapon states, and randomize the interval.
        if(longInterval.intervalElapsed()){
            //Refresh the target
            target = TargetManager(motherShip,target,false);
            strafeDistance = BestStrafeDistance(ship,true);
            longInterval.randomize();
        }

        //CONTINUAL CONTROL STATES
        //Checks booleans and gives continual commands here.
        //Done here so that we can use IntervalUtil to save CPU later
        if(!stopAccelerating){
            if(isAccelerating){
                ship.giveCommand(ShipCommand.ACCELERATE,null,0);
            }else{
                ship.giveCommand(ShipCommand.ACCELERATE_BACKWARDS,null,0);
            }
        } else {
            ship.giveCommand(ShipCommand.DECELERATE,null,0);
        }
        //Turning and strafing.
        if(canCircleStrafe && shouldStrafe){
            if(strafeRight){
                ship.giveCommand(ShipCommand.STRAFE_RIGHT,null,0);
            }
            if(strafeLeft){
                ship.giveCommand(ShipCommand.STRAFE_LEFT,null,0);
            }
        }

        //Calls the ship's System AI and the Cloak AI and lets it do its thing, whatever that is.
        if(script != null){
            script.advance(amount,destination,destination,target);
        }
        if(cloakScript != null){
            cloakScript.advance(amount,destination,destination,target);
        }



        //MAIN AI LOGIC
        //This code governs the ship's behavior and outputs boolean states that...
        //...are used in the continual control state code.
        //By using an IntervalUtil here, we can do practical load-balancing at the cost of accuracy.
        //if(checkInterval.intervalElapsed()){
			//Sets up Autofire
            AutoFire(ship);

            //FLUX CONTROL
            //If Flux is greater than 75% of max, Vent if possible.
            if((ship.getFluxTracker().getHardFlux() >= ship.getFluxTracker().getMaxFlux() * 0.75f)
            || (ship.getFluxTracker().getCurrFlux() >= ship.getFluxTracker().getMaxFlux() * 0.9f))
            {
                ship.giveCommand(ShipCommand.VENT_FLUX,null,0);
            }
            ShieldManager(ship,target,destination);

            //UPDATE TARGETS
            //If we don't have a target, find one.
            //If our target is dead, removed from the sim or has switched sides, find a new target.
            if (    target == null // unset
                    ||  target.isHulk()
                    ||  (ship.getOwner() == target.getOwner())        // friendly
                    ||  !Global.getCombatEngine().isEntityInPlay(target) // completely removed
                    || MathUtils.getDistanceSquared(motherShip,target) > ((roamDistance + strafeDistance) * (roamDistance * strafeDistance))
                    )
            {
                target = TargetManager(motherShip,target,false);
                if(target != null){
                    ship.setShipTarget(target);
                } else {
                    ship.setShipTarget(null);
                }
            }

            //DESTINATION CONTROL
            //Drones always use their mother ship as the destination
            destination = motherShip.getLocation();
            if(motherShip.isHulk() || motherShip == null){
                if(target != null){
                    destination = target.getLocation();
                }
            }

		switch (droneSystem.getDroneOrders()) {
			case ATTACK:
				//ROAM
				if(target != null && MathUtils.getDistanceSquared(ship.getLocation(),target.getLocation()) < ((roamDistance * roamDistance) - (target.getCollisionRadius() * target.getCollisionRadius())) ){
					shouldStrafe = true;
					steerMe(target.getLocation(),true,false);
				} else {
					shouldStrafe = false;
					steerMe(motherShip.getLocation(),false,true);
				}	break;
			case DEPLOY:
				//DEPLOY
				if(target != null && MathUtils.getDistanceSquared(ship.getLocation(),target.getLocation()) < ((closeDistance * closeDistance) - (target.getCollisionRadius() * target.getCollisionRadius())) ){
					shouldStrafe = true;
					steerMe(target.getLocation(),true,false);
				} else {
					shouldStrafe = false;
					steerMe(motherShip.getLocation(),false,true);
				}	break;
			default:
				shouldStrafe = false;
				steerMe(droneSystem.getLandingLocation(ship),false,true);
				destination = droneSystem.getLandingLocation(ship);
				break;
		}

		if(droneSystem.getDroneOrders().equals(DroneOrders.RECALL) && MathUtils.getDistanceSquared(ship.getLocation(),droneSystem.getLandingLocation(ship)) < (64f * 64f)){
			ship.beginLandingAnimation(ship);
		}

    }

    private void steerMe(Vector2f TLoc, boolean kiteable, boolean steerHome){
      //Get ship's velocity and location
      Vector2f MVel = ship.getVelocity();
      Vector2f MLoc = ship.getLocation();
      float MFace = ship.getFacing();

      //Main aiming logic
      float AngleToEnemy = VectorUtils.getAngle(MLoc, TLoc);
      //How closely aligned are we with the target?
      float angleToTarget = getAngleDifference(MFace, AngleToEnemy);

      float AbsAngD = Math.abs(angleToTarget);

      //////////////////////////////////////////////////////////////////////////////////////////////
      //Strafe a little bit to correct our aim.
      if (AbsAngD > 5 && MathUtils.getDistanceSquared(TLoc,MLoc) <= (strafeDistance * strafeDistance) && kiteable && !steerHome)
      {
          //course correct for ship velocity vector (some bugs when severely off target)
          float MFlightAng = VectorUtils.getAngle(new Vector2f(0, 0), MVel);
          float MFlightCC = getAngleDifference(MFace, MFlightAng);
          if (Math.abs(MFlightCC)>20)
          {
              if(MFlightCC < 0){
                  strafeLeft = true;
                  strafeRight = false;
                  //kiteable = false;
              } else {
                  strafeRight = true;
                  strafeLeft = false;
                  //kiteable = false;
              }
          }
      }

      //point towards target, perfectly.  Necessary with Drones, otherwise they get too fussy about steeting
        if (AbsAngD >= 0.5)
        {
            if(angleToTarget > 0){
                ship.setAngularVelocity(AbsAngD * 15);
            }else{
                ship.setAngularVelocity(AbsAngD * -15);
            }
        }

        //Force the angle to be perfect when close; prevents landing screwups.
        if(AbsAngD < 15){
            ship.setAngularVelocity(0);
            ship.setFacing(VectorUtils.getAngle(MLoc,TLoc));
        }

      //Acceleration code; if we're close to our target and we're facing it and it's suitable for kiting, back up and kite a bit.
      if(!steerHome){
        if (kiteable && shouldStrafe && MathUtils.getDistanceSquared(TLoc,MLoc) <= (strafeDistance * strafeDistance))
        {
            canCircleStrafe = true;
            isAccelerating = false;
            stopAccelerating = false;
        }else {
            canCircleStrafe = false;
            isAccelerating = true;
            stopAccelerating = false;
        }
      } else {
        canCircleStrafe = false;
        isAccelerating = true;
        stopAccelerating = false;
      }
      //Final stages; get whether ship is Overloaded or Venting and back up while wiggling if necessary.
      FluxTrackerAPI tracker = ship.getFluxTracker();
      if(tracker.isOverloadedOrVenting()){
           canCircleStrafe = true;
          isAccelerating = false;
      }
    }

    //Can be called outside the AI
    @Override
    public void setDoNotFireDelay(float amount) {
    }

    //Only called when Player puts ship back under AI control
    @Override
    public void forceCircumstanceEvaluation() {
        //target = findBestTarget(ship);
    }

    //Only gets set via other scripts here, but can be caught by other AI scripts (carriers or healing drones, for example).
    @Override
    public boolean needsRefit() {
        return false;
    }

	@Override
	public ShipwideAIFlags getAIFlags() {
		return flags;
	}

	@Override
	public void cancelCurrentManeuver() {
	}
}