package data.scripts.shipai;

import com.fs.starfarer.api.Global;
import com.fs.starfarer.api.combat.BattleCreationContext;
import com.fs.starfarer.api.combat.BattleObjectiveAPI;
import com.fs.starfarer.api.combat.CombatAssignmentType;
import com.fs.starfarer.api.combat.CombatFleetManagerAPI;
import com.fs.starfarer.api.combat.CombatFleetManagerAPI.AssignmentInfo;
import com.fs.starfarer.api.combat.CombatTaskManagerAPI;
import com.fs.starfarer.api.combat.ShipAIPlugin;
import com.fs.starfarer.api.combat.ShipAPI;
import com.fs.starfarer.api.combat.ShipCommand;
import com.fs.starfarer.api.combat.ShipSystemAIScript;
import com.fs.starfarer.api.combat.ShipwideAIFlags;
import com.fs.starfarer.api.fleet.FleetGoal;
import com.fs.starfarer.api.loading.WingRole;
import com.fs.starfarer.api.util.IntervalUtil;

import static data.scripts.plugins.ShipThreatPlugin.getCloseThreatsIncludeFighters;
import static data.scripts.plugins.ShipThreatPlugin.getFriendlies;
import static data.scripts.plugins.ShipThreatPlugin.getMostDangerousThreat;
import static data.scripts.plugins.ShipThreatPlugin.getNearestEnemyToPoint;
import static data.scripts.plugins.ShipThreatPlugin.getNearestFriendlyToPoint;
import static data.scripts.plugins.ShipThreatPlugin.getPlayerShip;

import static data.scripts.shipai.AutoFireController.AutoFire;
import static data.scripts.shipai.FluxManager.FluxManager;
import static data.scripts.shipai.ShieldManager.ShieldManager;
import static data.scripts.shipai.SteerMeStandard.SteerMeStandard;
import static data.scripts.shipai.TargetManager.TargetManager;
import static data.scripts.shipai.ThreatChecker.ThreatChecker;

import java.util.List;

import org.lazywizard.lazylib.MathUtils;

import org.lwjgl.util.vector.Vector2f;
import static data.scripts.shipai.BestStrafeDistance.BestStrafeDistance;

//Frigate AI
//Attempts to engage other Frigates and weak ships first.
public class StandardShipAI implements ShipAIPlugin
{
    // Our ship object
    private final ShipAPI ship;

    //Our System AI
    private final ShipSystemAIScript script;
    //Phase Cloak uses its own sub-AI for convenience.
    private final ShipSystemAIScript cloakScript;

    // Our current target (can be null)
    private ShipAPI target = null;

	private boolean isUnderOrders = false;

	private boolean isAccelerating = true;
	private boolean turnLeft = false;
	private boolean turnRight = false;

    private boolean strafeRight = false;
    private boolean strafeLeft = false;
    //Used for steering decisions
    private Vector2f destination = null;
    private Vector2f destinationVel = null;
	private Vector2f defaultDestination = null;

	//Our short decision-matrix stuff here.  Anything here must be very efficient code.
    private final IntervalUtil checkInterval = new IntervalUtil(0.01f, 0.05f);
	private int frameCounter = 0;
    //longInterval causes a new evalutation of targets regardless of anything else, to prevent AIs getting stuck stupidly.
    //Warning; don't run this too often, otherwise the AI won't even aim the guns properly.
	//Some very expensive code runs here, btw; don't up the interval a lot or you'll have issues w/ performance.
    private final IntervalUtil longInterval = new IntervalUtil(3f, 5f);
	//A very long interval clock that resets the position a ship seeks if it doesn't have a valid target and other long-term tasks.
	//This would be a good place for very expensive code, like strategic threat analysis, etc.
	private final IntervalUtil veryLongInterval = new IntervalUtil(15f, 30f);

    private final boolean shouldStrafe;
    private float strafeDistance;
    private boolean canCircleStrafe = false;
    private boolean strafeState = true;
    private final IntervalUtil strafeInterval = new IntervalUtil(0.5f,2.5f);
	private final ShipwideAIFlags flags = new ShipwideAIFlags();
	private boolean isTryingToVent = false;
	private boolean escorting = false;

    //This is populated by the SteerMe() routine
    private boolean[] commandList = new boolean[6];

    //Initial setup.  If waitTime = 0f, the guidance begins one frame after launch.
    //Otherwise, drifts until waitTime = 0f (useful for anime-style launch systems)
    public StandardShipAI (ShipAPI theShip, ShipSystemAIScript theScript, ShipSystemAIScript theCloakScript, boolean theShouldStrafe, float theStrafeDistance)
    {
        this.ship = theShip;
        this.shouldStrafe = theShouldStrafe;
        this.strafeDistance = theStrafeDistance;
        this.script = theScript;
		//Creates safe flags for this ship so that AIs won't crash, hopefully.
		for(ShipwideAIFlags.AIFlags flag : ShipwideAIFlags.AIFlags.values()){
			ship.getAIFlags().setFlag(flag);
			ship.getAIFlags().unsetFlag(flag);
		}
        if(script != null && ship.getSystem() != null){
            script.init(ship,ship.getSystem(),ship.getAIFlags(),Global.getCombatEngine());
        }
        this.cloakScript = theCloakScript;
        if(cloakScript != null && ship.getPhaseCloak() != null){
            cloakScript.init(ship,ship.getPhaseCloak(),null,Global.getCombatEngine());
        }
		ship.setPullBackFighters(false);
    }

    //Main logic
    @Override
    public void advance(float amount)
    {
        //LOGIC CHECKS
        if(Global.getCombatEngine() == null) return;
        if (Global.getCombatEngine().isPaused()) return;
		if(!Global.getCombatEngine().isInCampaign() && !Global.getCombatEngine().isInCampaignSim() && !Global.getCombatEngine().isMission() && !Global.getCombatEngine().isSimulation()) return;

        //If the engine's not paused, tick tock
        checkInterval.advance(amount);
        longInterval.advance(amount);
		veryLongInterval.advance(amount);
        strafeInterval.advance(amount);

		frameCounter += 1;
		if(frameCounter > 5){
			frameCounter = 0;
		}

        //Automatically strafe back and forth when that's vaguely appropriate.
        if(strafeInterval.intervalElapsed()){
            if(canCircleStrafe && shouldStrafe){
                strafeState = !strafeState;
                if(strafeState){
                    strafeRight = true;
                    strafeLeft = false;
                    strafeInterval.randomize();
                }else{
                    strafeRight = false;
                    strafeLeft = true;
                    strafeInterval.randomize();
                }
            }
        }

        //If enough time has passed, re-evaluate our best target, reset our weapon states, and randomize the interval.
        if(longInterval.intervalElapsed() || target == null){
            //Refresh the target
            target = TargetManager(ship,target,isTryingToVent);
            if(target != null){
                ship.setShipTarget(target);
            } else {
                ship.setShipTarget(null);
            }
            strafeDistance = BestStrafeDistance(ship,isUnderOrders);
            longInterval.randomize();
        }

		if(veryLongInterval.intervalElapsed() || defaultDestination == null){
			List<BattleObjectiveAPI> objList = Global.getCombatEngine().getObjectives();
			if(objList != null){
				if(!objList.isEmpty()){
					defaultDestination = MathUtils.getRandomPointInCircle(objList.get((int)MathUtils.getRandomNumberInRange(0,objList.size()-1)).getLocation(),2500f);
				} else {
					defaultDestination = MathUtils.getRandomPointInCircle(new Vector2f(0f,0f), 10000f);
				}
			} else {
				defaultDestination = MathUtils.getRandomPointInCircle(new Vector2f(0f,0f), 10000f);
			}
			//Explicitly sets / unsets the Vanilla AIFlags here
			for(ShipwideAIFlags.AIFlags flag : ShipwideAIFlags.AIFlags.values()){
				ship.getAIFlags().setFlag(flag);
				ship.getAIFlags().unsetFlag(flag);
			}
			veryLongInterval.randomize();
		}

        //CONTINUAL CONTROL STATES
        //Checks booleans and gives continual commands here.
        //Done here so that we can use IntervalUtil to save CPU later
        if(isAccelerating){
            ship.giveCommand(ShipCommand.ACCELERATE,null,0);
        }else{
			ship.giveCommand(ShipCommand.ACCELERATE_BACKWARDS,null,0);
        }
        //Turning and strafing.
        if(turnRight){
            ship.giveCommand(ShipCommand.TURN_RIGHT,null,0);
        }
        if (turnLeft){
            ship.giveCommand(ShipCommand.TURN_LEFT,null,0);
        }
        if(strafeRight){
            ship.giveCommand(ShipCommand.STRAFE_RIGHT,null,0);
        }
        if(strafeLeft){
            ship.giveCommand(ShipCommand.STRAFE_LEFT,null,0);
        }

        //Calls the ship's System AI and the Cloak AI and lets it do its thing, whatever that is.
        if(script != null){
            script.advance(amount,destination,destination,target);
        }
        if(cloakScript != null){
            cloakScript.advance(amount,destination,destination,target);
        }




        //MAIN AI LOGIC
        //This code governs the ship's behavior and outputs boolean states that...
        //...are used in the continual control state code.
        //By using an IntervalUtil here, we can do practical load-balancing at the cost of accuracy.
        if(checkInterval.intervalElapsed()){
			//Resets the target if the Player has recently intervened
			if(ship.isRecentlyShotByPlayer() && !ship.isAlly()){
				ShipAPI possPlayer = getPlayerShip();
				if(possPlayer != null){
					target = possPlayer;
				}
			}

            //Sets up Autofire
            AutoFire(ship);

			escorting = false;

			//Flux management system; manages Venting and triggers AI behaviors
			isTryingToVent = false;
			isTryingToVent = FluxManager(ship,target);

			//Manages our shield, if any.
            ShieldManager(ship,target,destination);

			//AI THREAT LOGIC
			if(!isTryingToVent && !ship.isFighter()){
				isTryingToVent = !escorting && ThreatChecker(ship,getCloseThreatsIncludeFighters(ship),getFriendlies(ship),isUnderOrders);

				if(isTryingToVent){
					//Set the target to the most-dangerous threat, so that we're backing away from it.
					ShipAPI posTarg = getMostDangerousThreat(ship,getCloseThreatsIncludeFighters(ship));
					if(posTarg != null && !posTarg.isAlly() && !posTarg.isHulk()){
						if(MathUtils.getDistanceSquared(ship, posTarg) <= (1500f * 1500f))target = posTarg;
					}
				}
			}

			if(isTryingToVent){
				ship.getAIFlags().setFlag(ShipwideAIFlags.AIFlags.IN_CRITICAL_DPS_DANGER);
				ship.getAIFlags().setFlag(ShipwideAIFlags.AIFlags.BACKING_OFF);
				ship.getAIFlags().setFlag(ShipwideAIFlags.AIFlags.BACK_OFF);
				ship.getAIFlags().setFlag(ShipwideAIFlags.AIFlags.TURN_QUICKLY);
			}


            //UPDATE TARGETS
            //If we don't have a target, find one.
            //If our target is dead, removed from the sim or has switched sides, find a new target.
            if (        target == null // unset
                    ||  target.isHulk()
					||	!target.isAlive()
					||	target.isAlly() //Ally
                    ||  (ship.getOwner() == target.getOwner())        // friendly
                    ||  !Global.getCombatEngine().isEntityInPlay(target) // completely removed
                    )
            {
                target = null;
                target = TargetManager(ship,target,isTryingToVent);
                if(target != null){
                    ship.setShipTarget(target);
                } else {
                    ship.setShipTarget(null);
                }
            }

            //Handles retreating case
            if(ship.isRetreating()){
                target = null;
                ship.setShipTarget(null);
            }

            //Destination MUST be reset to null at this stage of the state machine!
            destination = null;
            destinationVel = null;
			
			//Fighter controls; fighters / drones now obey carrier commands, sorta
			if(ship.isFighter() || ship.isDrone()){
				if(ship.isFighter() && (ship.getWing().getRole().equals(WingRole.ASSAULT) || ship.getWing().getRole().equals(WingRole.SUPPORT) || ship.getWing().getRole().equals(WingRole.BOMBER))){
					
					ShipAPI carrier = ship.getWing().getSourceShip();
					if(carrier != null && carrier.isAlive() && !carrier.isHulk()){
						target = carrier.getShipTarget();
						if(target != null){
							destination = target.getLocation();
						}
						if(carrier.isPullBackFighters()){
							destination = carrier.getLocation();
						}
					}
				} else {
					ShipAPI carrier = ship.getDroneSource();
					if(carrier != null && carrier.isAlive() && !carrier.isHulk()){
						target = carrier.getShipTarget();
						if(target != null){
							destination = target.getLocation();
						}
						if(carrier.isPullBackFighters()){
							destination = carrier.getLocation();
						}						
					}					
				}
			} else {
				//DESTINATION CONTROL
				//Does this ship have a destination of some kind?  Yes?  Good, go there!
				CombatFleetManagerAPI fleetManager = Global.getCombatEngine().getFleetManager(ship.getOwner());

				//Player fleets use the standard Command Point system.
				if(fleetManager != null){
					CombatTaskManagerAPI CTM = fleetManager.getTaskManager(false);
					AssignmentInfo info = CTM.getAssignmentFor(ship);
					Vector2f targetLoc = new Vector2f(0f,0f);
					if (info != null && info.getType() != null && info.getTarget() != null && info.getTarget().getLocation() != null){
						targetLoc = info.getTarget().getLocation();
					}
					CombatAssignmentType assignmentType;
					if(info != null && info.getType() != null){
						assignmentType = info.getType();
						if(assignmentType == null){
						   destination = null;
						   destinationVel = null;
						} else if (	assignmentType == CombatAssignmentType.ASSAULT ||
									assignmentType == CombatAssignmentType.CAPTURE ||
									assignmentType == CombatAssignmentType.CONTROL ||
									assignmentType == CombatAssignmentType.DEFEND ||
									assignmentType == CombatAssignmentType.RECON ||
									assignmentType == CombatAssignmentType.REPAIR_AND_REFIT ||
									assignmentType == CombatAssignmentType.RALLY_CARRIER ||
									assignmentType == CombatAssignmentType.RALLY_CIVILIAN ||
									assignmentType == CombatAssignmentType.RALLY_FIGHTERS ||
									assignmentType == CombatAssignmentType.RALLY_STRIKE_FORCE ||
									assignmentType == CombatAssignmentType.RALLY_TASK_FORCE
								) {
							ship.getAIFlags().setFlag(ShipwideAIFlags.AIFlags.PURSUING);
							ship.getAIFlags().setFlag(ShipwideAIFlags.AIFlags.MANEUVER_TARGET);
							ship.getAIFlags().setFlag(ShipwideAIFlags.AIFlags.RUN_QUICKLY);
							isUnderOrders = false;
							destination = targetLoc;
							destinationVel = new Vector2f(0,0);
						} else if(assignmentType == CombatAssignmentType.SEARCH_AND_DESTROY){
							ship.getAIFlags().setFlag(ShipwideAIFlags.AIFlags.PURSUING);
							ship.getAIFlags().setFlag(ShipwideAIFlags.AIFlags.MANEUVER_TARGET);
							ship.getAIFlags().setFlag(ShipwideAIFlags.AIFlags.RUN_QUICKLY);
							isUnderOrders = false;
							destination = null;
							destinationVel = null;
						} else if (assignmentType == CombatAssignmentType.RETREAT){
						   BattleCreationContext context =  Global.getCombatEngine().getContext();
						   FleetGoal playerGoal = context.getPlayerGoal();
						   FleetGoal otherGoal = context.getOtherGoal();
							ship.getAIFlags().setFlag(ShipwideAIFlags.AIFlags.NEEDS_HELP);
							if(ship.getOwner() == 0){
								if(playerGoal == FleetGoal.ESCAPE){
									destination = new Vector2f(ship.getLocation().getX(),ship.getLocation().getY() + 50000f);
									destinationVel = new Vector2f();
									target = null;
								}else {
									destination = new Vector2f(ship.getLocation().getX(),ship.getLocation().getY() - 50000f);
									destinationVel = new Vector2f();
									target = null;
								}
							} else {
								if(playerGoal != FleetGoal.ESCAPE && otherGoal == FleetGoal.ESCAPE){
									destination = new Vector2f(ship.getLocation().getX(),ship.getLocation().getY() + 50000f);
									destinationVel = new Vector2f();
									target = null;
								}else {
									destination = new Vector2f(ship.getLocation().getX(),ship.getLocation().getY() + 50000f);
									destinationVel = new Vector2f();
									target = null;
								}
							}
						} else if ( assignmentType == CombatAssignmentType.LIGHT_ESCORT ||
									assignmentType == CombatAssignmentType.MEDIUM_ESCORT ||
									assignmentType == CombatAssignmentType.HEAVY_ESCORT
									){
							isUnderOrders = true;
							escorting = true;
							//If given a direct order to follow a ship, and that ship has a target, then attack it, if equal or smaller size.
							ShipAPI escorted = getNearestFriendlyToPoint(ship,info.getTarget().getLocation());
							if(escorted != null){
								if(escorted.getShipTarget() != null  && !escorted.getShipTarget().isHulk() && Global.getCombatEngine().isEntityInPlay(escorted.getShipTarget()) && MathUtils.getDistance(escorted,escorted.getShipTarget()) < 2500f){
									destination = escorted.getLocation();
									destinationVel = escorted.getVelocity();
									target = escorted.getShipTarget();
								} else {
									destination = escorted.getLocation();
									destinationVel = escorted.getVelocity();
								}
							} else {
								destination = targetLoc;
								destinationVel = info.getTarget().getVelocity();
							}
						} else if (assignmentType == CombatAssignmentType.ENGAGE
									|| assignmentType == CombatAssignmentType.HARASS
									|| assignmentType == CombatAssignmentType.INTERCEPT
									|| assignmentType == CombatAssignmentType.STRIKE
								){
							ship.getAIFlags().setFlag(ShipwideAIFlags.AIFlags.PURSUING);
							ship.getAIFlags().setFlag(ShipwideAIFlags.AIFlags.MANEUVER_TARGET);
							ship.getAIFlags().setFlag(ShipwideAIFlags.AIFlags.RUN_QUICKLY);
							isUnderOrders = true;
							//If given a direct order to go kill a Thing, go kill that Thing, do not pass Go, do not collect $200.
							destination = targetLoc;
							destinationVel = info.getTarget().getVelocity();
							if (destination == null || destinationVel == null){
								return;
							}
							ShipAPI closeBy = getNearestEnemyToPoint(ship,destination);
							if(closeBy != null){
								target = closeBy;
							}
						} else if(info.getTarget() != null){
							//Catch-all for weird orders here.
							ship.getAIFlags().setFlag(ShipwideAIFlags.AIFlags.PURSUING);
							ship.getAIFlags().setFlag(ShipwideAIFlags.AIFlags.MANEUVER_TARGET);
							ship.getAIFlags().setFlag(ShipwideAIFlags.AIFlags.RUN_QUICKLY);
							isUnderOrders = false;
							destination = targetLoc;
							destinationVel = info.getTarget().getVelocity();
						} else{
							//If all else fails, let the AI just figure it out.
							ship.getAIFlags().setFlag(ShipwideAIFlags.AIFlags.PURSUING);
							ship.getAIFlags().setFlag(ShipwideAIFlags.AIFlags.MANEUVER_TARGET);
							ship.getAIFlags().setFlag(ShipwideAIFlags.AIFlags.RUN_QUICKLY);
							isUnderOrders = false;
							destination = null;
							destinationVel = null;
						}
					}
				} else if (fleetManager == null) {
					//If all else fails, let the AI just figure it out.
					isUnderOrders = false;
					destination = null;
					destinationVel = null;
				}
			}
            //Sets a combat destination using a randomized circle for attacking distant targets who aren't too fast for this to matter
            //The two distance settings (for the outside distance and the circle position) are crucial and a bit picky!
			if(target!=null){
				boolean canKite = true;
				if(ship.isFighter() && ship.getWing().getRole().equals(WingRole.BOMBER)){
					canKite = false;
				}
				if(ship.isFighter() && (target.isFighter() || target.isDrone())){
					canKite = false;
				}
				if(destination != null){
					if(escorting && MathUtils.getDistanceSquared(ship, target) < (1000f * 1000f) && MathUtils.getDistanceSquared(ship, destination) < (1000f * 1000f)){
						//If the target is close, go ahead and steer to it if we need to.
						commandList = SteerMeStandard(ship, target, strafeDistance, target.getLocation(), target.getVelocity(), canKite, shouldStrafe, isTryingToVent);
					} else if (escorting && MathUtils.getDistanceSquared(ship, destination) >= (500f * 500f)){
						//Go straight to the ecorted ship.
						commandList = SteerMeStandard(ship, null, 0f, destination, new Vector2f(0f,0f), false, false, false);
					}else if(MathUtils.getDistanceSquared(ship, target) < (1000f * 1000f) && MathUtils.getDistanceSquared(ship, destination) < (1000f * 1000f)){
						//If the target is close, go ahead and steer to it if we need to.
						commandList = SteerMeStandard(ship, target, strafeDistance, target.getLocation(), target.getVelocity(), canKite, shouldStrafe, isTryingToVent);
					}else if(MathUtils.getDistanceSquared(ship, destination) < (2000f * 2000f)){
						//If we're fairly close to the destination, go ahead and engage
						commandList = SteerMeStandard(ship, target, strafeDistance, target.getLocation(), target.getVelocity(), canKite, shouldStrafe, isTryingToVent);
					} else {
						//Go straight to the destination.
						commandList = SteerMeStandard(ship, null, 0f, destination, new Vector2f(0f,0f), false, false, false);
					}
				} else {
					if(SteerMeStandard.isBiggerOrEqual(target,ship) && MathUtils.getDistanceSquared(ship, target) > (2500f * 2500f)){
						//If we can, get behind the target
						Vector2f targetRear = MathUtils.getPointOnCircumference(target.getLocation(), 2000f, MathUtils.clampAngle(target.getFacing() -180f));
						commandList = SteerMeStandard(ship, target, strafeDistance, targetRear, target.getVelocity(), canKite, shouldStrafe, isTryingToVent);
					} else {
						//Else, don't bother, go straight there, and back up if trying to Vent, etc.
						commandList = SteerMeStandard(ship, target, strafeDistance, target.getLocation(), target.getVelocity(), canKite, shouldStrafe, isTryingToVent);
					}
				}
			} else if (destination != null){
				//Go straight to the destination.
				commandList = SteerMeStandard(ship, null, 0f, destination, new Vector2f(0f,0f), false, false, false);
			} else {
				//Nothing?  Well, go somewhere vaguely useful.  Vaguely.
				commandList = SteerMeStandard(ship, null, 0f, defaultDestination, new Vector2f(0f,0f), false, false, false);
			}

            turnLeft = commandList[0];
            turnRight = commandList[1];
            strafeLeft = commandList[2];
            strafeRight = commandList[3];
            canCircleStrafe = commandList[4];
            isAccelerating = commandList[5];
        }
    }

    //Can be called outside the AI
    @Override
    public void setDoNotFireDelay(float amount) {
    }

    //Only called when Player puts ship back under AI control
    @Override
    public void forceCircumstanceEvaluation() {
        /*target = TargetManager(ship,target,isTryingToVent);
        if(target != null){
            ship.setShipTarget(target);
        }*/
    }

    //Only gets set via other scripts here, but can be caught by other AI scripts (carriers or healing drones, for example).
    @Override
    public boolean needsRefit() {
        return false;
    }

	@Override
	public ShipwideAIFlags getAIFlags() {
		return flags;
	}

	@Override
	public void cancelCurrentManeuver() {
		isUnderOrders = false;
		destination = null;
		destinationVel = null;
		target = null;
	}
}