package data.scripts.weapons.ai;
import com.fs.starfarer.api.Global;
import com.fs.starfarer.api.combat.AutofireAIPlugin;
import com.fs.starfarer.api.combat.CombatEntityAPI;
import com.fs.starfarer.api.combat.ShipAPI;
import com.fs.starfarer.api.combat.WeaponAPI;
import com.fs.starfarer.api.util.IntervalUtil;
import data.scripts.plugins.ShipThreatPlugin;
import data.scripts.weapons.helperUtils.VacuumWeaponUtils;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import org.lazywizard.lazylib.CollisionUtils;
import org.lazywizard.lazylib.MathUtils;
import org.lazywizard.lazylib.VectorUtils;
import org.lwjgl.util.vector.Vector2f;

public class BoardingShuttleAutoFireAI implements AutofireAIPlugin {
    
    private float theWeaponAngle;
    private Vector2f theWeaponLoc;
    private float theWeaponArc;
    private float theWeaponRange;
    private final WeaponAPI theWeapon;
    private final float barrelLength;
    private Vector2f theTargetLoc;
    private ShipAPI manTarget;
    private float energyCost = -100000;
    private final float tolerance;
    private boolean isBeam = false;
    private boolean shouldFire = false;
    private boolean holdFire = true;
    private CombatEntityAPI theTarget = null;
    private final IntervalUtil checkInterval = new IntervalUtil(0.20f, 0.30f);
    
    private static final Set IGNORE_IDS = new HashSet();
    static
    {
        // Add all projectiles that should pierce shields here
        IGNORE_IDS.add("neutral_swift");
        IGNORE_IDS.add("neutral_jean_station");
        IGNORE_IDS.add("neutral_marlene_station");
    }    
    
    //Initial setup.  If waitTime = 0f, the guidance begins one frame after launch.  
    //Otherwise, drifts until waitTime = 0f (useful for anime-style launch systems)
    public BoardingShuttleAutoFireAI (WeaponAPI weapon, float theBarrelLength, float theTolerance)
    {
        this.theWeapon = weapon;
        this.barrelLength = theBarrelLength;
        this.tolerance = theTolerance;
    }  
    
    private List<ShipAPI> getEnemiesInArcWithTolerance(WeaponAPI weapon)
    {
        List<ShipAPI> enemies = new ArrayList<>();
         //Gets all nearby ships that may be valid threats.
        List<ShipAPI> threats = ShipThreatPlugin.getHulks();
        if(threats == null || threats.isEmpty()) return null;          
        
        for (ShipAPI ship : threats)
        {
            if (
                    !ship.isDrone()
                    && !ship.isFighter()
                    && !IGNORE_IDS.contains(ship.getHullSpec().getHullId())
                    && ship.getOriginalOwner() != weapon.getShip().getOriginalOwner()
                    && VacuumWeaponUtils.vacuumIsWithinArc(ship,theWeapon,theWeaponLoc,theWeaponRange,theWeaponArc,theWeaponAngle) 
                ) 
            {
                enemies.add(ship);
            }
        }
        return enemies;
    }     
    
    private List<ShipAPI> getEnemiesInArcLong(WeaponAPI weapon)
    {
        List<ShipAPI> enemies = new ArrayList<>();
        //Gets all nearby ships that may be valid threats.
        List<ShipAPI> threats = ShipThreatPlugin.getThreats(weapon.getShip());
        if(threats == null || threats.isEmpty()) return null;           
        
        for (ShipAPI ship : threats)
        {
            if (    
                    ship.isHulk()
                    && !ship.isDrone()
                    && !ship.isFighter()
                    && ship.getOriginalOwner() != weapon.getShip().getOriginalOwner()
                    && !IGNORE_IDS.contains(ship.getHullSpec().getHullId())
                    && VacuumWeaponUtils.vacuumIsWithinArc(ship,theWeapon,theWeaponLoc,theWeaponRange,theWeaponArc,theWeaponAngle)
                )
            {
                enemies.add(ship);
            }
        }
        return enemies;
    }      
    
    //This is our sorter, that finds the closest ShipAPI that is in the arc that is not a Drone.
    public CombatEntityAPI findBestTarget(WeaponAPI weapon) 
    {
        holdFire = false;
        ShipAPI source = weapon.getShip();
        if(source != null){
            manTarget = source.getShipTarget();
            if( 
                    manTarget != null 
                    &&  manTarget.isHulk() 
                    &&  !(source.getOwner() == manTarget.getOwner())    // not friendly
                    && VacuumWeaponUtils.vacuumIsWithinArc(manTarget,theWeapon,theWeaponLoc,theWeaponRange,theWeaponArc,theWeaponAngle)
                )
            {
                holdFire = false;
                return manTarget;
            }
        }     
 
        List<ShipAPI> nearbyEnemies = getEnemiesInArcWithTolerance(weapon);
        if(nearbyEnemies != null && !nearbyEnemies.isEmpty()){
            for (ShipAPI enemy : nearbyEnemies) {
                holdFire = false;
                return enemy;
            }
        }
        
         //Still no targets?  Look long and see what's available
        holdFire = true;
        nearbyEnemies = getEnemiesInArcLong(weapon);
        if(nearbyEnemies != null && !nearbyEnemies.isEmpty()){
            return (CombatEntityAPI) nearbyEnemies.get(0);
        }
        return null;
    }    
    
    private Vector2f getFiringSolution(Vector2f source, Vector2f destLocation, Vector2f destVelocity, float projVel)
    {
        float tx = destLocation.getX() - source.getX();
        float ty = destLocation.getY() - source.getY();
        float tvx = destVelocity.getX();
        float tvy = destVelocity.getY();

        // Get quadratic equation components
        float a = tvx*tvx + tvy*tvy - projVel*projVel;
        float b = 2 * (tvx * tx + tvy * ty);
        float c = tx*tx + ty*ty;    

        // Solve quadratic
        Vector2f ts = quad(a, b, c); // See quad(), below

        // Find smallest positive solution
        Vector2f finalSolution = new Vector2f();
        if(ts != null) 
        {
            float t0 = ts.getX();
            float t1 = ts.getY();
            float t = Math.min(t0, t1);
            if (t < 0) t = Math.max(t0, t1);    
            if (t > 0) 
            {
                finalSolution.x = destLocation.getX() + destVelocity.getX()*t;
                finalSolution.y = destLocation.getY() + destVelocity.getY() * t;
            }
        }
        return finalSolution; 
    }
    
    private Vector2f quad(float a, float b,float c) {
      Vector2f sol = new Vector2f();
      if (Math.abs(a) < 1e-6) {
            if (Math.abs(b) < 1e-6) {
                if(Math.abs(c) < 1e-6){
                    sol.set(0f,0f);
                } else {
                    sol = null;
                }
            } else {
                sol.x = -c / b;
                sol.y = -c / b;
            }
        } else {
            float disc = b*b - 4*a*c;
            if (disc >= 0f) {
                disc = (float) Math.sqrt(disc);
                a = 2f*a;
                sol.set((-b-disc)/a, (-b+disc)/a);
            }
      }
      return sol;
    } 
    
    // Will be in next LazyLib version
    public static float getAngleDifference(float angle1, float angle2)
    {
        float distance = (angle2 - angle1) + 180f;
        distance = (distance / 360.0f);
        distance = ((distance - (float) Math.floor(distance)) * 360f) - 180f;
        return distance;
    }    

    @Override
    public void advance(float amount) {
        
        if(Global.getCombatEngine() == null) return;
        if (Global.getCombatEngine().isPaused()) return;
        //No ammo?  Don't bother.
        if(theWeapon.usesAmmo() && theWeapon.getAmmo() == 0){
            shouldFire = false;
            return;
        }        
        
         //Sets up the points for our arc-check optimization
        theWeaponArc = theWeapon.getArc() + (tolerance * 2f);
        theWeaponLoc = theWeapon.getLocation();
        theWeaponAngle = theWeapon.getCurrAngle();
        theWeaponRange = barrelLength + theWeapon.getRange();        

        //If neither of the above are true, go ahead and increment our checkInterval
        checkInterval.advance(amount);
        if(checkInterval.intervalElapsed()){
            if(theTarget == null){
                theTarget = findBestTarget(theWeapon);//Every once in awhile, refresh the target (gets manual targets)
            }
            checkInterval.randomize();
            return;
        }
        //Set up Flux variables for later.
        if(energyCost == -100000){
            isBeam = theWeapon.isBeam() || theWeapon.isBurstBeam();
            energyCost = theWeapon.getFluxCostToFire();
            if(energyCost < 1.0f){
                energyCost = 0.0f;
            }
        }

        //UPDATE TARGETS
        //If we don't have a target, stop.   
        //If our target is dead, out of range or removed from the sim or has switched sides, find a new target.
        if (    
                theTarget == null // unset
                ||  !Global.getCombatEngine().isEntityInPlay(theTarget) //Completely removed from play
                ||  barrelLength == -500f  //catches screwed-up barrelLength
                ||  holdFire  //We have a distant target, but we need to update and not fire     
                ||  ((theWeapon.getShip().getFluxTracker().getCurrFlux() + energyCost) > theWeapon.getShip().getFluxTracker().getMaxFlux()) //No flux
                ||  !VacuumWeaponUtils.vacuumIsWithinArc(theTarget,theWeapon,theWeaponLoc,theWeaponRange,theWeaponArc,theWeaponAngle)
            ) 
        {
            shouldFire = false;
            theTarget = null;
            return;
        } 
        //If our target is a ShipAPI, don't shoot at it if it's cloaked this frame.
        if(theTarget instanceof ShipAPI){
            ShipAPI theShip = (ShipAPI) theTarget;
            if(theShip.getPhaseCloak() != null){
               if((theShip.getPhaseCloak().isOn() || theShip.getPhaseCloak().isActive()) && theShip.getPhaseCloak().getId().contains("phasecloak")){
                   theTargetLoc = theTarget.getLocation();
                   shouldFire = false;
                   theTarget = null;
                   return;
               }
           }           
        }

        if(theTarget != null && !holdFire){
            if(isBeam){
                theTargetLoc = theTarget.getLocation();               
                Vector2f offset = MathUtils.getPointOnCircumference(theWeaponLoc,theWeaponRange,theWeapon.getCurrAngle());   
                
                shouldFire = CollisionUtils.getCollides(theWeaponLoc,
                                offset,
                                theTarget.getLocation(),
                                theTarget.getCollisionRadius());
            } else{
                //For projectile guns, getting the correct point to shoot at is considerably more complicated than for laser beams (duh!)
                //To do this, we need to take the barrel's length into account (at least right now) and we also need to factor the relative velocity of
                //the shooter and target.  The first part of this code rotates a line segment so that we have the right offset from the barrel
                //once we have that and combVelocity (our relative velocity) we can then do the quadratic equation needed to get a valid aiming point.
                Vector2f offset = MathUtils.getPointOnCircumference(theWeaponLoc,barrelLength,theWeapon.getCurrAngle());
                
                Vector2f combVelocity = new Vector2f(theTarget.getVelocity().getX() - theWeapon.getShip().getVelocity().getX(),theTarget.getVelocity().getY() -theWeapon.getShip().getVelocity().getY());
                
                theTargetLoc = getFiringSolution(offset,theTarget.getLocation(),combVelocity,theWeapon.getProjectileSpeed());
                //Are we aimed at the target?
                float AngleToEnemy = VectorUtils.getAngle(theWeaponLoc, theTargetLoc);
                float angleToTarget = getAngleDifference(theWeapon.getCurrAngle(), AngleToEnemy);
                float AbsAngD = Math.abs(angleToTarget);
                shouldFire = AbsAngD - tolerance < 1f;//tighter for non-beams, since they need to lead well
            }   
        } else {
            shouldFire = false;
        }
    }

    @Override
    public boolean shouldFire() {
        return shouldFire;
    }

    @Override
    public void forceOff() {
        //nothing
    }

    @Override
    public Vector2f getTarget() {
        if(theTarget != null){
            return theTargetLoc;
        }
        return null;
    }

    @Override
    public ShipAPI getTargetShip() {
        if(theTarget instanceof ShipAPI) return (ShipAPI) theTarget;            
        return null;
    }

    @Override
    public WeaponAPI getWeapon() {
        return theWeapon;
    }

}

