package data.scripts.weapons.ai;
import com.fs.starfarer.api.Global;
import com.fs.starfarer.api.combat.AutofireAIPlugin;
import com.fs.starfarer.api.combat.CombatEntityAPI;
import com.fs.starfarer.api.combat.ShipAPI;
import com.fs.starfarer.api.combat.WeaponAPI;
import com.fs.starfarer.api.util.IntervalUtil;
import java.util.ArrayList;
import java.util.List;
import org.lazywizard.lazylib.CollisionUtils;
import org.lazywizard.lazylib.MathUtils;
import org.lazywizard.lazylib.VectorUtils;
import org.lazywizard.lazylib.combat.WeaponUtils;
import org.lwjgl.util.vector.Vector2f;

public class RepairDroneAutoFireAI implements AutofireAIPlugin {
    
    private final WeaponAPI theWeapon;
    private final float barrelLength;
    private Vector2f theTargetLoc;
    private ShipAPI manTarget;
    private float energyCost = -100000;
    private final float tolerance;
    private boolean isBeam = false;
    private boolean shouldFire = false;
    private boolean holdFire = true;
    private CombatEntityAPI theTarget = null;
    private final IntervalUtil checkInterval = new IntervalUtil(0.20f, 0.30f);
    
    //Initial setup.  If waitTime = 0f, the guidance begins one frame after launch.  
    //Otherwise, drifts until waitTime = 0f (useful for anime-style launch systems)
    public RepairDroneAutoFireAI (WeaponAPI weapon, float theBarrelLength, float theTolerance)
    {
        this.theWeapon = weapon;
        this.barrelLength = theBarrelLength;
        this.tolerance = theTolerance;
    }  
    
    private List<ShipAPI> getFriendliesInArcWithTolerance(WeaponAPI weapon)
    {
        int weaponOwner = weapon.getShip().getOwner();
        List<ShipAPI> friends = new ArrayList<>();
        float range = weapon.getRange();

        for (ShipAPI ship : Global.getCombatEngine().getShips())
        {
            float hitpoints = ship.getHitpoints();
            float maxHitpoints = ship.getMaxHitpoints();
            if (    
                    //Get friendlies.
                    ship.getOwner() == weaponOwner
                    && hitpoints < maxHitpoints 
                    && !ship.isDrone()
                    && !ship.isHulk()
                    && weapon.distanceFromArc(ship.getLocation()) - tolerance <= 0f
                    && MathUtils.isWithinRange(ship, weapon.getLocation(), range)
                ) 
            {
                if(ship.getPhaseCloak() != null){
                    if(!ship.getPhaseCloak().isOn() && !ship.getPhaseCloak().isActive() && !ship.getPhaseCloak().getId().contains("phasecloak")){
                        friends.add(ship);
                    }
                } else {
                    friends.add(ship);
                }
            }
        }
        return friends;
    }     
    
    private List<ShipAPI> getFriendliesInArcLong(WeaponAPI weapon)
    {
        int weaponOwner = weapon.getShip().getOwner();
        List<ShipAPI> friends = new ArrayList<>();
        float range = weapon.getRange() * 2f;

        for (ShipAPI ship : Global.getCombatEngine().getShips())
        {
            float hitpoints = ship.getHitpoints();
            float maxHitpoints = ship.getMaxHitpoints();           
            if (    
                    //Get friendlies.
                    ship.getOwner() == weaponOwner
                    && hitpoints < maxHitpoints
                    && !ship.isDrone()
                    && !ship.isHulk()
                    && weapon.distanceFromArc(ship.getLocation()) - tolerance <= 0f
                    && MathUtils.isWithinRange(ship, weapon.getLocation(), range)
                )
            {
                if(ship.getPhaseCloak() != null){
                    if(!ship.getPhaseCloak().isOn() && !ship.getPhaseCloak().isActive() && !ship.getPhaseCloak().getId().contains("phasecloak")){
                        friends.add(ship);
                    }
                } else {
                    friends.add(ship);
                }
            }
        }
        return friends;
    } 
    
    //This is our sorter, that finds the closest ShipAPI that is in the arc that is not a Drone.
    public CombatEntityAPI findBestTarget(WeaponAPI weapon) 
    {
        holdFire = false;
        ShipAPI source = weapon.getShip();
        if(source != null){
            manTarget = source.getShipTarget();
            if( 
                    manTarget != null 
                    &&  !manTarget.isHulk() 
                    &&  !(source.getOwner() == manTarget.getOwner())    // not friendly
                    &&  Global.getCombatEngine().isEntityInPlay(manTarget) //Still exists
                    &&  MathUtils.isWithinRange(manTarget, theWeapon.getLocation(), theWeapon.getRange())
                    &&  WeaponUtils.isWithinArc(manTarget,theWeapon) // can be engaged
                )
            {
                holdFire = false;
                return manTarget;
            }
        }     

        //Get nearby friends
        List<ShipAPI> nearbyFriends = getFriendliesInArcWithTolerance(weapon);
        for (ShipAPI friend : nearbyFriends) {
            holdFire = false;
            return friend;
        }
		
        //Still no targets?  Look long and see what's available
        holdFire = true;
        nearbyFriends = getFriendliesInArcLong(weapon);
        if(!nearbyFriends.isEmpty()){
            return (CombatEntityAPI) nearbyFriends.get(0);
        }
        return null;
    }    
    
    private Vector2f getFiringSolution(Vector2f source, Vector2f destLocation, Vector2f destVelocity, float projVel)
    {
        float tx = destLocation.getX() - source.getX();
        float ty = destLocation.getY() - source.getY();
        float tvx = destVelocity.getX();
        float tvy = destVelocity.getY();

        // Get quadratic equation components
        float a = tvx*tvx + tvy*tvy - projVel*projVel;
        float b = 2 * (tvx * tx + tvy * ty);
        float c = tx*tx + ty*ty;    

        // Solve quadratic
        Vector2f ts = quad(a, b, c); // See quad(), below

        // Find smallest positive solution
        Vector2f finalSolution = new Vector2f();
        if(ts != null) 
        {
            float t0 = ts.getX();
            float t1 = ts.getY();
            float t = Math.min(t0, t1);
            if (t < 0) t = Math.max(t0, t1);    
            if (t > 0) 
            {
                finalSolution.x = destLocation.getX() + destVelocity.getX()*t;
                finalSolution.y = destLocation.getY() + destVelocity.getY() * t;
            }
        }
        return finalSolution; 
    }
    
    private Vector2f quad(float a, float b,float c) {
      Vector2f sol = new Vector2f();
      if (Math.abs(a) < 1e-6) {
            if (Math.abs(b) < 1e-6) {
                if(Math.abs(c) < 1e-6){
                    sol.set(0f,0f);
                } else {
                    sol = null;
                }
            } else {
                sol.x = -c / b;
                sol.y = -c / b;
            }
        } else {
            float disc = b*b - 4*a*c;
            if (disc >= 0f) {
                disc = (float) Math.sqrt(disc);
                a = 2f*a;
                sol.set((-b-disc)/a, (-b+disc)/a);
            }
      }
      return sol;
    } 
    
    // Will be in next LazyLib version
    public static float getAngleDifference(float angle1, float angle2)
    {
        float distance = (angle2 - angle1) + 180f;
        distance = (distance / 360.0f);
        distance = ((distance - (float) Math.floor(distance)) * 360f) - 180f;
        return distance;
    }    

    @Override
    public void advance(float amount) {
        
        if(Global.getCombatEngine() == null) return;
        if (Global.getCombatEngine().isPaused()) return;
        //No ammo?  Don't bother.
        if(theWeapon.usesAmmo() && theWeapon.getAmmo() == 0){
            shouldFire = false;
            return;
        }        

        //If neither of the above are true, go ahead and increment our checkInterval
        checkInterval.advance(amount);
        if(checkInterval.intervalElapsed()){
            if(theTarget == null){
                theTarget = findBestTarget(theWeapon);//Every once in awhile, refresh the target (gets manual targets)
            }
            checkInterval.randomize();
            return;
        }
        //Set up Flux variables for later.
        if(energyCost == -100000){
            isBeam = theWeapon.isBeam() || theWeapon.isBurstBeam();
            energyCost = theWeapon.getFluxCostToFire();
            if(energyCost < 1.0f){
                energyCost = 0.0f;
            }
        }

        //UPDATE TARGETS
        //If we don't have a target, stop.   
        //If our target is dead, out of range or removed from the sim or has switched sides, find a new target.
        if (    
                theTarget == null // unset 
                ||  (theWeapon.getShip().getOwner() != theTarget.getOwner()) //Enemy!
                ||  !Global.getCombatEngine().isEntityInPlay(theTarget) //Completely removed from play
                ||  barrelLength == -500f  //catches screwed-up barrelLength
                ||  holdFire  //We have a distant target, but we need to update and not fire     
                ||  ((theWeapon.getShip().getFluxTracker().getCurrFlux() + energyCost) > theWeapon.getShip().getFluxTracker().getMaxFlux()) //No flux
                ||  !WeaponUtils.isWithinArc(theTarget,theWeapon) // Has left the arc
                ||  !MathUtils.isWithinRange(theTarget, theWeapon.getLocation(), theWeapon.getRange() * 1.1f) //Not in range
            ) 
        {
            shouldFire = false;
            theTarget = null;
            return;
        } 
        //If our target is a ShipAPI, don't shoot at it if it's cloaked this frame.
        if(theTarget instanceof ShipAPI){
            ShipAPI theShip = (ShipAPI) theTarget;
            if(theShip.isHulk()){
                   shouldFire = false;
                   theTarget = null;
                   return;                
            }            
            if(theShip.getPhaseCloak() != null){
               if((theShip.getPhaseCloak().isOn() || theShip.getPhaseCloak().isActive()) && theShip.getPhaseCloak().getId().contains("phasecloak")){
                   theTargetLoc = theTarget.getLocation();
                   shouldFire = false;
                   theTarget = null;
                   return;
               }
           }           
        }

        if(theTarget != null && !holdFire){
            if(isBeam){
                theTargetLoc = theTarget.getLocation();               
                /*
                double[] pt = {(double) barrelLength + theWeapon.getRange(), (double) 0f};
                AffineTransform.getRotateInstance(Math.toRadians(theWeapon.getCurrAngle()), 0, 0).transform(pt, 0, pt, 0, 1); // specifying to use this double[] to hold coords
                float newX = (float) pt[0];
                float newY = (float) pt[1];
                Vector2f offset = new Vector2f(theWeapon.getLocation().getX() + newX, theWeapon.getLocation().getY() + newY);  */
                Vector2f offset = MathUtils.getPointOnCircumference(theWeapon.getLocation(),barrelLength + theWeapon.getRange(),theWeapon.getCurrAngle());   
                
                shouldFire = CollisionUtils.getCollides(theWeapon.getLocation(),
                                offset,
                                theTarget.getLocation(),
                                theTarget.getCollisionRadius());
            } else{
                //For projectile guns, getting the correct point to shoot at is considerably more complicated than for laser beams (duh!)
                //To do this, we need to take the barrel's length into account (at least right now) and we also need to factor the relative velocity of
                //the shooter and target.  The first part of this code rotates a line segment so that we have the right offset from the barrel
                //once we have that and combVelocity (our relative velocity) we can then do the quadratic equation needed to get a valid aiming point.
                //Many thanks to Louis Wasserman for this example of simple line segment rotation on Stack Overflow.
                /*
                double[] pt = {(double) barrelLength, (double) 0f};
                AffineTransform.getRotateInstance(Math.toRadians(theWeapon.getCurrAngle()), 0, 0).transform(pt, 0, pt, 0, 1); // specifying to use this double[] to hold coords
                float newX = (float) pt[0];
                float newY = (float) pt[1];
                Vector2f offset = new Vector2f(theWeapon.getLocation().getX() + newX, theWeapon.getLocation().getY() + newY);*/
                Vector2f offset = MathUtils.getPointOnCircumference(theWeapon.getLocation(),barrelLength,theWeapon.getCurrAngle());
                
                Vector2f combVelocity = new Vector2f(theTarget.getVelocity().getX() - theWeapon.getShip().getVelocity().getX(),theTarget.getVelocity().getY() -theWeapon.getShip().getVelocity().getY());
                
                theTargetLoc = getFiringSolution(offset,theTarget.getLocation(),combVelocity,theWeapon.getProjectileSpeed());
                //Are we aimed at the target?
                float AngleToEnemy = VectorUtils.getAngle(theWeapon.getLocation(), theTargetLoc);
                float angleToTarget = getAngleDifference(theWeapon.getCurrAngle(), AngleToEnemy);
                float AbsAngD = Math.abs(angleToTarget);
                shouldFire = AbsAngD - tolerance < 1f;//tighter for non-beams, since they need to lead well
            }   
        } else {
            shouldFire = false;
        }
    }

    @Override
    public boolean shouldFire() {
        return shouldFire;
    }

    @Override
    public void forceOff() {
        //nothing
    }

    @Override
    public Vector2f getTarget() {
        if(theTarget != null){
            return theTargetLoc;
        }
        return null;
    }

    @Override
    public ShipAPI getTargetShip() {
        if(theTarget instanceof ShipAPI) return (ShipAPI) theTarget;            
        return null;
    }

    @Override
    public WeaponAPI getWeapon() {
        return theWeapon;
    }

}

