package data.scripts.weapons.ai;
import com.fs.starfarer.api.Global;
import com.fs.starfarer.api.combat.AutofireAIPlugin;
import com.fs.starfarer.api.combat.CollisionClass;
import com.fs.starfarer.api.combat.CombatEntityAPI;
import com.fs.starfarer.api.combat.ShipAPI;
import com.fs.starfarer.api.combat.ShipwideAIFlags;
import com.fs.starfarer.api.combat.WeaponAPI;
import com.fs.starfarer.api.util.IntervalUtil;
import data.scripts.plugins.ShipThreatPlugin;
import static data.scripts.weapons.ai.WeaponAIUtilities.friendlyFireCheck;
import static data.scripts.weapons.ai.WeaponAIUtilities.getAngleDifference;
import static data.scripts.weapons.ai.WeaponAIUtilities.getFiringSolution;
import data.scripts.weapons.helperUtils.VacuumWeaponUtils;
import java.util.ArrayList;
import java.util.List;
import org.lazywizard.lazylib.CollisionUtils;
import org.lazywizard.lazylib.MathUtils;
import org.lazywizard.lazylib.VectorUtils;
import org.lwjgl.util.vector.Vector2f;

public class StandardAutoFireAI implements AutofireAIPlugin {

    private float theWeaponAngle;
    private Vector2f theWeaponLoc;
    private final boolean isHardpoint;
    private float theWeaponArc;
    private float theWeaponRange;
    private final WeaponAPI theWeapon;
    private final float barrelLength;
    private Vector2f theTargetLoc;
    private ShipAPI manTarget;
    private final float tolerance;
    private final boolean isBeam;
    private boolean shouldFire = false;
    private boolean holdFire = true;
    private CombatEntityAPI theTarget = null;
    private final IntervalUtil checkInterval = new IntervalUtil(0.20f, 0.30f);
    
    //Initial setup.  If waitTime = 0f, the guidance begins one frame after launch.  
    //Otherwise, drifts until waitTime = 0f (useful for anime-style launch systems)
    public StandardAutoFireAI (WeaponAPI weapon, float theBarrelLength, float theTolerance)
    {
        theWeapon = weapon;
        barrelLength = theBarrelLength;
        tolerance = theTolerance;       
		isBeam = theWeapon.isBeam() || theWeapon.isBurstBeam();
        isHardpoint = theWeapon.getSlot().isHardpoint();
    }  
    
    private List<ShipAPI> getEnemiesInArcWithTolerance(WeaponAPI weapon)
    {
        //Empty array that gets populated with possible targets later on.
        List<ShipAPI> enemies = new ArrayList<>();

        //Gets all nearby ships that may be valid threats.
        List<ShipAPI> threats = ShipThreatPlugin.getCloseThreatsIncludeFighters(weapon.getShip());
		if(weapon.getShip().isDrone()){
			if(weapon.getShip().getDroneSource() != null){
				threats = ShipThreatPlugin.getCloseThreatsIncludeFighters(weapon.getShip().getDroneSource());
			}
		}	
        if(threats == null || threats.isEmpty()) return null;
        for (ShipAPI ship : threats)
        {   
            if(VacuumWeaponUtils.vacuumIsWithinArc(ship,theWeapon,theWeaponLoc,theWeaponRange,theWeaponArc,theWeaponAngle)){
                enemies.add(ship);
            }
        }
        return enemies;
    }     
    
    private List<ShipAPI> getEnemiesInArcLong(WeaponAPI weapon)
    {
        //Empty array that gets populated with possible targets later on.
        List<ShipAPI> enemies = new ArrayList<>();

        //Gets all nearby ships that may be valid threats.
        List<ShipAPI> threats = ShipThreatPlugin.getCloseThreatsIncludeFighters(weapon.getShip());
		if(weapon.getShip().isDrone()){
			if(weapon.getShip().getDroneSource() != null){
				threats = ShipThreatPlugin.getCloseThreatsIncludeFighters(weapon.getShip().getDroneSource());
			}
		}			
        if(threats == null || threats.isEmpty()) return null;
        for (ShipAPI ship : threats)
        {
            if(VacuumWeaponUtils.vacuumIsWithinArc(ship,theWeapon,theWeaponLoc,theWeaponRange * 1.5f,theWeaponArc,theWeaponAngle)){
                enemies.add(ship);
            }
        }
        return enemies;
    } 
    
    //This is our sorter, that finds the closest ShipAPI that is in the arc that is not a Drone.
    private CombatEntityAPI findBestTarget(WeaponAPI weapon) 
    {
        holdFire = false;
        ShipAPI source = weapon.getShip();
            
        if(source != null){
            manTarget = source.getShipTarget();
            if( 
                    manTarget != null 
                    &&  !manTarget.isHulk() 
                    &&  source.getOwner() != manTarget.getOwner()    // not friendly
                    && VacuumWeaponUtils.vacuumIsWithinArc(manTarget,theWeapon,theWeaponLoc,theWeaponRange,theWeaponArc,theWeaponAngle)
                )
            {
                holdFire = false;
                return manTarget;
            }
        }     

        //Let's get all of the ShipAPIs in the weapon's arc, sorted by distance 
        List<ShipAPI> nearbyEnemies = getEnemiesInArcWithTolerance(theWeapon);
        
        if(nearbyEnemies != null){
            for (ShipAPI enemy : nearbyEnemies) {
                //criteria to add to list (you can put anything that will return a true/false here)
                if (!enemy.isDrone() && !enemy.isFighter())
                {
                    holdFire = false;
                    return enemy;
                }
            }
            for (ShipAPI enemy : nearbyEnemies) {
                //criteria to add to list (you can put anything that will return a true/false here)
                if (!enemy.isDrone())
                {
                    holdFire = false;
                    return enemy;
                }
            }			
        }
        
        //Add Drones back in, if we haven't found anything better.
        if(nearbyEnemies != null && !nearbyEnemies.isEmpty()){
            holdFire = false;
            return nearbyEnemies.get(0);
        }
        
        //Still no targets?  Look long and see what's available
        holdFire = true;
        nearbyEnemies = getEnemiesInArcLong(theWeapon);
        if(nearbyEnemies != null && !nearbyEnemies.isEmpty()){
            return (CombatEntityAPI) nearbyEnemies.get(0);
        }
        return null;
    }   

    @Override
    public void advance(float amount) {        
        shouldFire = false;
        if(Global.getCombatEngine() == null) return;
        if (Global.getCombatEngine().isPaused()) return;
        
        //No ammo?  Don't bother.
        if(theWeapon.usesAmmo() && theWeapon.getAmmo() == 0){
            shouldFire = false;
            return;
        } 
		
		//Quits firing if the ship's Flux levels are too high.
		if(theWeapon.getShip().getAIFlags() != null){
			if(theWeapon.usesEnergy() && theWeapon.getShip().getAIFlags().hasFlag(ShipwideAIFlags.AIFlags.DO_NOT_AUTOFIRE_NON_ESSENTIAL_GROUPS)){
				shouldFire = false;
				return;
			}	
		}
        
        //Sets up the points for our arc-check optimization; each step here needs to be done once a frame (fairly expensive but unavoidable)
        theWeaponArc = theWeapon.getArc() + (tolerance * 2f);
        theWeaponLoc = theWeapon.getLocation();
        theWeaponAngle = theWeapon.getCurrAngle();
        theWeaponRange = barrelLength + theWeapon.getRange();
        
        if(theTarget != null 
			&& !theWeapon.getProjectileCollisionClass().equals(CollisionClass.PROJECTILE_NO_FF)
			&& !theWeapon.getProjectileCollisionClass().equals(CollisionClass.MISSILE_NO_FF)
			){
            shouldFire = friendlyFireCheck(theWeapon.getShip(),theWeaponLoc,theWeaponRange,theWeaponAngle);
            if(shouldFire == false){
                return;
            }
        }  

        //If neither of the above are true, go ahead and increment our checkInterval
        checkInterval.advance(amount);
        if(checkInterval.intervalElapsed()){
            theTarget = findBestTarget(theWeapon);//Every once in awhile, refresh the target (gets manual targets)
            checkInterval.randomize();
        }

        //UPDATE TARGETS
        //If we don't have a target, stop.   
        //If our target is dead, out of range or removed from the sim or has switched sides, find a new target.
        if (    
                theTarget == null // unset 
                //||  (theWeapon.getShip().getOwner() == theTarget.getOwner()) //Friendly!
                //||  barrelLength == -500f  //catches screwed-up barrelLength
                ||  holdFire  //We have a distant target, but we need to update and not fire     
                ||  !VacuumWeaponUtils.vacuumIsWithinArc(theTarget,theWeapon,theWeaponLoc,theWeaponRange,theWeaponArc,theWeaponAngle)
            )
        {
            shouldFire = false;
            theTarget = null;
            theTarget = findBestTarget(theWeapon);
            if(theTarget != null){
                theTargetLoc = theTarget.getLocation();
            }
            return;
        } 
        
        //If our target is a ShipAPI, don't shoot at it if it's cloaked this frame.
        if(theTarget instanceof ShipAPI){
            ShipAPI theShip = (ShipAPI) theTarget;
            if(theShip.isHulk()){
                shouldFire = false;
                theTarget = null;
                theTarget = findBestTarget(theWeapon);
                return;                
            }
            if(theShip.getPhaseCloak() != null){
               if((theShip.getPhaseCloak().isOn() || theShip.getPhaseCloak().isActive()) && theShip.getPhaseCloak().getId().contains("phasecloak")){
                   theTargetLoc = theTarget.getLocation();
                   shouldFire = false;
                   theTarget = null;
                   theTarget = findBestTarget(theWeapon);
                   return;
               }
           }           
        }

        if(theTarget != null && !holdFire){
            if(isBeam){
                theTargetLoc = theTarget.getLocation();               
                Vector2f offset = MathUtils.getPointOnCircumference(theWeaponLoc,theWeaponRange,theWeaponAngle);   
                
                //First test, because it's cheaper.
                shouldFire = CollisionUtils.isPointWithinCollisionCircle(offset,theTarget);
                if(shouldFire == false){
                    shouldFire = CollisionUtils.getCollides(theWeaponLoc,
                                    offset,
                                    theTargetLoc,
                                    theTarget.getCollisionRadius());
                }
            } else {
                if(isHardpoint == true){
                    theTargetLoc = theTarget.getLocation();               
                    Vector2f offset = MathUtils.getPointOnCircumference(theWeaponLoc,theWeaponRange,theWeaponAngle);   

                    //First test, because it's cheaper.
                    shouldFire = CollisionUtils.isPointWithinCollisionCircle(offset,theTarget);
                    if(shouldFire == false){
                        shouldFire = CollisionUtils.getCollides(theWeaponLoc,
                                        offset,
                                        theTargetLoc,
                                        theTarget.getCollisionRadius());
                    }
                } else {
                    //For projectile guns, getting the correct point to shoot at is considerably more complicated than for laser beams (duh!)
                    //To do this, we need to take the barrel's length into account and we also need to factor the relative velocity of
                    //the shooter and target.  The first part of this code rotates a line segment so that we have the right offset from the barrel
                    //once we have that and combVelocity (our relative velocity) we can then do the quadratic equation needed to get a valid aiming point.
                    //Many thanks to Louis Wasserman for this example of simple line segment rotation on Stack Overflow and to LazyWizard for an optimization.
                    Vector2f offset = MathUtils.getPointOnCircumference(theWeaponLoc,barrelLength,theWeaponAngle);

                    Vector2f combVelocity = new Vector2f(theTarget.getVelocity().getX() - theWeapon.getShip().getVelocity().getX(),theTarget.getVelocity().getY() -theWeapon.getShip().getVelocity().getY());

                    theTargetLoc = getFiringSolution(offset,theTarget.getLocation(),combVelocity,theWeapon.getProjectileSpeed());
                    //Are we aimed at the target?
                    float AngleToEnemy = VectorUtils.getAngle(theWeaponLoc, theTargetLoc);
                    float angleToTarget = getAngleDifference(theWeaponAngle, AngleToEnemy);
                    float AbsAngD = Math.abs(angleToTarget);
                    shouldFire = AbsAngD - tolerance < 1f;//tighter for non-beams, since they need to lead well
                }
            }   
        } else {
            if(theTarget != null){
                theTargetLoc = theTarget.getLocation();
            }
            shouldFire = false;
        }
    }

    @Override
    public boolean shouldFire() {
        return shouldFire;
    }

    @Override
    public void forceOff() {
        //nothing
    }

    @Override
    public Vector2f getTarget() {
        if(theTarget != null){
            return theTargetLoc;
        }
        return null;
    }

    @Override
    public ShipAPI getTargetShip() {
        if(theTarget instanceof ShipAPI) return (ShipAPI) theTarget;            
        return null;
    }

    @Override
    public WeaponAPI getWeapon() {
        return theWeapon;
    }
}

