package data.scripts.weapons.missileai;

import com.fs.starfarer.api.Global;
import com.fs.starfarer.api.combat.CombatEntityAPI;
import com.fs.starfarer.api.combat.MissileAIPlugin;
import com.fs.starfarer.api.combat.MissileAPI;
import com.fs.starfarer.api.combat.ShipAPI;
import com.fs.starfarer.api.combat.ShipCommand;
import com.fs.starfarer.api.util.IntervalUtil;
import data.scripts.plugins.ShipThreatPlugin;
import static data.scripts.weapons.ai.WeaponAIUtilities.getAngleDifference;
import static data.scripts.weapons.missileai.SteerMissileStandard.SteerMissileStandard;
import java.util.List;

import org.lazywizard.lazylib.VectorUtils;

import org.lwjgl.util.vector.Vector2f;

//GuidedMissileAI_NoDrones script
//A low-latency guided missile AI with an example of sorting out non-drone targets
//Based on Dmaiski's work on LRPD AI
public class GuidedMissileAI_NoDrones implements MissileAIPlugin
{
    // Our missile object
    private final MissileAPI missile;
    private float waitTime = 0f;
	private int spawnCount = 0;
	private float spawnDistance = 0f;
	private boolean isHeatSeeker = false;

    // Our current target (can be null)
    private CombatEntityAPI target = null;
    
    //Booleans set current control state, updated every tick
    private boolean isAccelerating = false;
    private boolean turnRight = false;
    private boolean turnLeft = false;
    private boolean strafeRight = false;
    private boolean strafeLeft = false;
	private boolean canStrafe = false;
	private boolean strafeDir = false;
        
    private IntervalUtil checkInterval = new IntervalUtil(0.05f, 0.1f);
	private IntervalUtil strafeInterval = new IntervalUtil(0.75f, 1.0f);
	
	//This is populated by the SteerMissileStandard() routine
    private boolean[] commandList = new boolean[6];
                
    //Initial setup.  If waitTime = 0f, the guidance begins one frame after launch.  
    //Otherwise, drifts until waitTime = 0f (useful for anime-style launch systems)
    public GuidedMissileAI_NoDrones (MissileAPI missile, float waitTime, int spawnC, float spawnDist, boolean heatSeek)
    {
        this.missile = missile;
        this.waitTime = waitTime;
        this.spawnCount = spawnC;
		this.spawnDistance = spawnDist;
		this.isHeatSeeker = heatSeek;
        //This line can be enabled to give missile targetting data even while drifting.
        //target = findBestTarget(missile);
    }
    
    //This is our sorter, that finds the closest ShipAPI that is not a Drone.
    private ShipAPI findBestTarget(MissileAPI missile) 
    {
        ShipAPI source = missile.getSource();
        if(source != null){
            ShipAPI manTarget = source.getShipTarget();
            if(manTarget != null && !manTarget.isHulk() &&  !(missile.getOwner() == manTarget.getOwner())        // friendly
                && Global.getCombatEngine().isEntityInPlay(manTarget))
            {
                return manTarget;
            }
        }
        //this is where your missile is
        Vector2f loc = missile.getLocation();

        //Get all enemies within range.
        if(source == null) return null;
        List<ShipAPI> nearbyEnemies = ShipThreatPlugin.getThreats(source);
        if(nearbyEnemies == null || nearbyEnemies.isEmpty()) return null;

        //If we don't have a manual target, then the missile looks for the closest one in the forward arc.
        float searchArc = 45f;
        for (ShipAPI enemy : nearbyEnemies) {
            if (!enemy.isDrone())
            {
                Vector2f TLoc = enemy.getLocation();
                float MFace = missile.getFacing();
                
                //main aiming (to find angle you are off target by)
                float AngleToEnemy = VectorUtils.getAngle(loc, TLoc);
                //Gets the difference in the angle to the enemy vs. the missile's current angle
                float angleToTarget = getAngleDifference(MFace, AngleToEnemy);
                float AbsAngD = Math.abs(angleToTarget);  
                if(AbsAngD < searchArc){
                    if(enemy.getPhaseCloak() != null){
                        if(!enemy.getPhaseCloak().isOn() || !enemy.getPhaseCloak().isActive()){
                            return enemy;
                        }
                    } else {
                        return enemy;
                    }
                }
            }
        }
        for (ShipAPI enemy : nearbyEnemies) {
            //Add in everything else, including Phased ships.  EXCLUDES DRONES.
            if(!enemy.isDrone() && !enemy.isHulk() && Global.getCombatEngine().isEntityInPlay(enemy)){
                return enemy;
            }            
        }

        return null;
    }

    //Main logic
    @Override
    public void advance(float amount)
    {        
        //LOGIC CHECKS
		//If paused or the combat engine isn't initialized, stop.
        if(Global.getCombatEngine() == null) return;
        if (Global.getCombatEngine().isPaused()) return;		
		
        // If fading or fizzling, don't bother with updates.
        if (missile.isFading() || missile.isFizzling()) return;
        
        //Do nothing while waitTime > 0
        if(waitTime > 0f){
            waitTime -= 0.16f;
            return;
        }
		
		if(canStrafe){
			strafeInterval.advance(amount);
			if(strafeInterval.intervalElapsed()){
				if(strafeDir){
					strafeRight = true;
					strafeLeft = false;
				} else {
					strafeRight = false;
					strafeLeft = true;
				}
				strafeInterval.randomize();
			}
		}
		
		//CONTINUAL CONTROL STATES
		//Checks booleans and gives continual commands here.
		//Done here so that we can use IntervalUtil to save CPU later
		if(isAccelerating){
			missile.giveCommand(ShipCommand.ACCELERATE);
		}else{
			missile.giveCommand(ShipCommand.DECELERATE);
		}
		if(turnRight){
			missile.giveCommand(ShipCommand.TURN_RIGHT);
		}
		if (turnLeft){
			missile.giveCommand(ShipCommand.TURN_LEFT);
		}
		if(strafeRight){
			missile.giveCommand(ShipCommand.STRAFE_RIGHT);
		}
		if(strafeLeft){
			missile.giveCommand(ShipCommand.STRAFE_LEFT);
		}		
        
        //If neither of the above are true, go ahead and increment our checkInterval
        checkInterval.advance(amount);
        
        //UPDATE TARGETS
        //If we don't have a target, find one.   
        //If our target is dead, removed from the sim or has switched sides, find a new target.
        if (target == null // unset 
                ||  (missile.getOwner() == target.getOwner())        // friendly
                ||  !Global.getCombatEngine().isEntityInPlay(target) ) // completely removed
        {
            target = findBestTarget(missile);
            return;
        }  
		if(target instanceof ShipAPI){
			ShipAPI targ = (ShipAPI) target;
			if(targ.isHulk()){
	            target = findBestTarget(missile);
				return;			
			}
		}
        
        //MAIN AI LOGIC
        //This code governs the missile's behavior and outputs boolean states that...
        //...are used in the continual control state code.
        //By using an IntervalUtil here, we can do practical load-balancing at the cost of accuracy.
        if (target!=null && checkInterval.intervalElapsed()) //if you have a target
        {
			//Can we spawn missiles from this missile?  Do so now.
			//This code is DISABLED until a future SS feature supports it.
			/*
			if(spawnCount > 0 && !missile.isFromMissile() && missile.getElapsed() > 0.5f){
				if(MathUtils.getDistanceSquared(missile, target) <= (spawnDistance * spawnDistance)){
					Global.getCombatEngine().applyDamage(missile, missile.getLocation(), 10000f, DamageType.HIGH_EXPLOSIVE, 0f, true, false, missile.getSource());
					for(int i = 0; i < spawnCount; i++){
						MissileAPI newProj = (MissileAPI) Global.getCombatEngine().spawnProjectile(
								missile.getSource(), 
								missile.getWeapon(), 
								missile.getWeapon().getId(), 
								missile.getLocation(), 
								MathUtils.getRandomNumberInRange(0f, 360f), 
								missile.getVelocity());
						newProj.setFromMissile(true);
						newProj.setMissileAI(new GuidedMissileAI_NoDrones(newProj,0.2f,0,0f));
					}
					return;
				}
			}*/		
			
            //Reset all of the continuous control states here.
			commandList = SteerMissileStandard(missile,target,isHeatSeeker);
			
			turnLeft = commandList[0];
			turnRight = commandList[1];
			strafeLeft = commandList[2];
			strafeRight = commandList[3];
			isAccelerating = commandList[4];	
			canStrafe = commandList[5];
			
			checkInterval.randomize();
        }
    }
}